/*
 *  ZMailer smtpserver child-registry
 *
 *  Registers/keeps track of IP addresses that are currently
 *  talking with us in order to figure out of there are too
 *  many parallel connections from same IP address out there..
 *
 */

#include "smtpserver.h"

int
childsameip(void*addr)
{
  return 0;
}

void
childregister(int pid, void *addr)
{
}

void
childreap(int pid)
{
}
